import 'dart:async';
import 'dart:io';
import 'dart:typed_data';

import 'package:flutter_xtransport/src/interface.dart';
import 'package:flutter_xtransport/src/jsons.dart';

class XTransportTcpClient implements XTransportClient {
  @override
  bool Function(X509Certificate)? onBadCertificate;
  final String _host;
  final int _port;
  Socket? _socket;

  bool busy = false;

  @override
  String? authority;
  @override
  SecurityContext? securityContext;
  @override
  void send(Packet obj) {
    _socket?.add(obj.pack());
  }

  RemoteInfo _remoteInfo = RemoteInfo();
  LocalInfo _localInfo = LocalInfo();
  StreamSubscription<Uint8List>? _sub;

  //Events
  void Function()? _onConnect;
  void Function()? _onClose;
  void Function(XTransportError err)? _onError;
  void Function(XMessage msg)? _onMessage;
  @override
  void onClose(void Function() fn) => _onClose = fn;

  @override
  void onConnect(void Function() fn) => _onConnect = fn;

  @override
  void onError(void Function(XTransportError err) fn) => _onError = fn;

  @override
  void onMessage(void Function(XMessage msg) fn) => _onMessage = fn;

  @override
  void close() {
    print("please close");
    _socket?.close();
    _socket?.destroy();
  }

  XTransportTcpClient(this._host, this._port,
      {this.authority, this.onBadCertificate, this.securityContext});

  Future<Socket> getConnectionSocket() async {
    // if (kDebugMode) {
    //   kde
    print("conecting...");
    // }
    var tmp =
        await Socket.connect(_host, _port, timeout: const Duration(seconds: 1));
    // Don't wait for io buffers to fill up before sending requests.
    if (tmp.address.type != InternetAddressType.unix) {
      tmp.setOption(SocketOption.tcpNoDelay, true);
    }

    if (securityContext != null) {
      // Todo(sigurdm): We want to pass supportedProtocols: ['h2'].
      // http://dartbug.com/37950
      // _low = tmp;
      return await SecureSocket.secure(
        tmp,
        host: authority ?? _host,
        context: securityContext,
        onBadCertificate: onBadCertificate,
      );
    }
    return tmp;
  }

  Future<void> connect() async {
    if (busy) {
      return Future.value();
    }
    busy = true;
    try {
      _socket = await getConnectionSocket();
      _remoteInfo = RemoteInfo(
        address: _socket?.remoteAddress.address ?? "",
        host: authority ?? _socket?.remoteAddress.host ?? "",
        port: _port,
        family: _socket?.remoteAddress.type.name ?? "",
      );
      _localInfo = LocalInfo(
        address: _localInfo.address,
        family: _socket?.address.type.name ?? "",
        port: _socket?.port ?? 0,
      );
      _onConnect?.call();
    } catch (e) {
      _onError?.call(XTransportError.from(e));
      busy = false;
      _onClose?.call();
      return Future.value();
    }
    broadcastNotifications();
    return Future.value();
  }

  Future<StreamSubscription<Uint8List>> broadcastNotifications() async {
    // _socket?.as
    var xx = _socket?.listen((streamData) {
      _onMessage?.call(XMessage(
          message: streamData,
          remoteInfo: RemoteInfo(
            address: _remoteInfo.address,
            host: _remoteInfo.host,
            port: _remoteInfo.port,
            family: _remoteInfo.family,
            size: streamData.length,
          ),
          localInfo: _localInfo));
    }, onDone: () {
      busy = false;
      _onClose?.call();
    }, onError: (e) {
      _onError?.call(XTransportError.from(e));
      _onClose?.call();
    }, cancelOnError: true);
    return Future.value(xx);
  }
}
